/* all X stuff in this file guesses based on 
    usr/doc/XFree86-doc-3.3.5/X11/xlib.PS.gz  


  TODO: 
    0) a) speed at expense of ram:
            keep an XImage representing the window in memory to be used
            to redraw on expose events (just redraw sections from this
            image ).  this eliminates slow scaling in current system.
            i.e. when placing tiles, place a copy to this XImage as well
            as to the screen.
       b) black line bug:
        instead of blitting tile to window/XImage, blit extracted section
        from out image with an extra pixel on the left (unless first tile)
    1)  make resizing work w/out segfaults
    2)  fill in gaps in scaled images
    3) segfaults on big images exposing; in bw_sVTC_down always
           -- i think fixed; was borderling case in x_sVTC_down; fixed both.


*/


#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>

#include "rpp.h"

#define DEBUG 0

void writeImagetoXImage( XImage *x, Image *i );

Display *display;
Visual *visual;
XPixmapFormatValues *XpmFormats;
int n_XpmFormats;
Window window;
GC gc;
double sx, sy; /* scale */
int win_w, win_h; /* last window dims (size of buf) */
//Pixmap p_buf;
//Pixmap p_tile;
int b_allocd = 0, t_allocd=0;

#define fmin(a,b) (a)<(b)?(a):(b)

/* connect to X server and setup w x h window */
void G_init(int w, int h){
  XSetWindowAttributes attribs;

  if(DEBUG){ printf("+1+"); fflush(stdout); }
  display = XOpenDisplay(NULL);
  if( display == NULL ) {
    printf("couldn't connect to X.\n");
    exit(-1);
  }
  if(DEBUG){ printf("+2+"); fflush(stdout); }
  visual = XDefaultVisual(display,0);
  gc = XDefaultGC(display, 0);

  /* scale window to fit display */
  //dh = XDisplayHeight( display, 0 ) - 20;
  //dw = XDisplayWidth( display, 0 ) - 20;

  /* ??? */
  XpmFormats = XListPixmapFormats( display, &n_XpmFormats );

  /* p. 31 */
  
  attribs.background_pixmap =  None;  /* background_pixmap */
  attribs.background_pixel =  XBlackPixel(display, 0); /* background pixel */
  attribs.border_pixmap = CopyFromParent; /* border pixmap */
  attribs.border_pixel =  XWhitePixel(display, 0); /* border pixel */
  attribs.bit_gravity =  ForgetGravity; /* bit gravity */
  attribs.win_gravity = NorthWestGravity; /* window gravity */
  attribs.backing_store = Always; /* backing store; save window's contents when obscured */
  attribs.backing_planes =  0;  /* backing store stuff; flag out */
  attribs.backing_pixel =  0;  /* " */
  attribs.save_under =  False; /* save stuff under this window */
  attribs.event_mask = ExposureMask; /* not interested in any events except expose, resize */
  attribs.do_not_propagate_mask = NoEventMask; /* propagate all events to ancestor ... (the root?) */
  attribs.override_redirect =  False; /* override redirect flag -- WM thing? */
  attribs.colormap =   CopyFromParent; /* color map */
  attribs.cursor =    None; /* cursor; inherits from parent */
   


  /* p.37 */
  window = XCreateWindow( display, XDefaultRootWindow(display), 10, 10, 
                           w, h, 0, CopyFromParent,
                           InputOutput, visual, 
			  CWBackPixmap | CWBackPixel | CWBorderPixmap |
                          CWBorderPixel | CWBitGravity | CWWinGravity |
                          CWBackingStore | CWOverrideRedirect | CWSaveUnder |
                          CWEventMask | CWDontPropagate | CWColormap | CWCursor
                           , 
                           &attribs);

  XMapWindow( display, window );
  win_w = w;
  win_h = h;

  if(DEBUG){ printf("+3+"); fflush(stdout); }
}




void G_end(void){
  /*
  XFree(stuff.... );
  .
  .
  . */
  /*
    if( b_allocd ){ 
      XFreePixmap( display, p_buf );
    }
    if( t_allocd ){
      XFreePixmap( display, p_tile );
    }
  */
  if(DEBUG){ printf("+A+"); fflush(stdout); }
  XUnmapWindow( display, window ); /* redundant? */
  if(DEBUG){ printf("+B+"); fflush(stdout); }
  XDestroyWindow( display, window );
  if(DEBUG){ printf("+C+"); fflush(stdout); }
  XCloseDisplay( display );

}


void G_place_pic( Image *tile, Image *out, int x, int y ){

  XImage p_tile;
  writeImagetoXImage( &p_tile, tile );
  placeRectinImg( out, tile, x, y );
  XInitImage( &p_tile );
  XPutImage( display,  window, gc, &p_tile, 0, 0, x , y, tile->w, tile->h );

  free( p_tile.data );
  

}

/* gfx functions - p. 138 */
/* gc functions - p.122 */
/* colour stuff - p.74 */
/* XImage stuff - p. 166 & p. 383 */

/* i'm amazed this works.  must be beginner's luck. */
void writeImagetoXImage( XImage *x, Image *i ){
  /* XColor c;  */ /* p. 75 */
  char *p, *q, *end;
 if(DEBUG){ printf("+m+"); fflush(stdout); }
  x->width = i->w;
  x->height = i->h;
  x->xoffset = 0;
  x->format = ZPixmap; /* scanline array of pixels */
  x->byte_order = MSBFirst;
  x->bitmap_unit = 32;
  x->bitmap_pad = 32;
  x->depth = 24; /* ??? */
  x->bytes_per_line = 4*x->width;
  x->bits_per_pixel = 32;
  x->red_mask = 255 << 16;
  x->green_mask = 255 << 8;
  x->blue_mask = 255;

 if(DEBUG){ printf("+n+"); fflush(stdout); }
  x->data = (char *)malloc( (i->w) * (i->h) * 32 * sizeof(char));
  end = i->data + (i->size)*i->bpp;
  p = i->data;
  q = x->data;
  if( i->bpp == 3 ){
    while( p < end ){
      *q++ = 0;  // byte order = XRGB
      *q++ = *p++;
      *q++ = *p++;
      *q++ = *p++;
    }
  } else if ( i->bpp == 1 ){
    while( p < end ){
      *q++ = 0;
      *q++ = *p;
      *q++ = *p;
      *q++ = *p++; 
    }
  } else {
    printf("%s: wrong bpp %d!\n", __FUNCTION__, i->bpp );
    exit(-1);
  }
 if(DEBUG){ printf("+o+"); fflush(stdout); }
  /*
  c.red = ((short )(*r)) << 8;
  c.green = ((short)(*g)) << 8;
  c.blue = ((short)(*b)) << 8;
  XQueryColor(display, ?colormap?, &c); */   /* p. 92 */


}


/* events - p.183 
   XExposeEvent def'd p. 200
   XResizeRequestEvent def'd p. 211 -- might fuck with WM.
   event handling - p.216
*/

void free_Image( Image *i ){
  if( i > 0 ){
    free(i->name);
    free(i->data);
    free(i);
  }
}



void processEvents( Image *out){
  XEvent e;
  Image *i;
  XImage x;


  
  while(XCheckWindowEvent(display, window, ExposureMask, &e)==True ){
    if( e.type == Expose ){
      // printf("expose: (%d,%d) (%d,%d) \n", e.xexpose.x, e.xexpose.y, e.xexpose.width, e.xexpose.height );
      i = extractRectFromImg( out, e.xexpose.x, e.xexpose.y, e.xexpose.width, e.xexpose.height );

      writeImagetoXImage( &x, i );
      free_Image(i);
      XInitImage( &x );
      XPutImage( display,  window, gc, &x, 0, 0, e.xexpose.x, e.xexpose.y, e.xexpose.width, e.xexpose.height );    
      free( x.data );

    } /* else if ( e.type == ResizeRequest ){
       
      printf("resizeRequest event recieved.\n");
      }*/
     else {
      printf("unhandled X event %d.\n", e.type );
      exit(-1);
    }    
  }


}
