#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <sys/types.h>
#include <signal.h>
#include <time.h>
#include <string.h>
#include <unistd.h>

#include "types.h"
#include "rpp.h"
#include "font.h"
#define POPSIZE 48
Image buf;
int compile = 2;

//#define efunc(x) double  x( int type, double value, int nkids, node **kids, double cx, double cy, double cz, double ct )
#define efunc(x) double x(int type, double value, int nkids, node **kids)

#define countArgs(x) if (nkids != x) { printf("Wrong # of args! \n"); return 0; }


//double evaluate( node *n, double cx, double cy, double cz, double ct);
//#define eval(x) evaluate(x, cx, cy, cz, ct)
// making parameters global for speed...
double evaluate( node *n);
#define eval(x) evaluate(x)
double ct, cz, cx,cy;

efunc(e_none){ return 0; }
efunc(e_num){ return value; }
efunc(e_x){ return cx; }
efunc(e_y){ return cy; }
efunc(e_z){ return cz; }
efunc(e_t){ return ct; }
efunc(e_mult){ 
  //countArgs(2)
  return eval(kids[0]) * eval(kids[1]);
}
efunc(e_add){
  //countArgs(2)
  return eval(kids[0]) + eval(kids[1]);
}
efunc(e_sub){
  countArgs(2)
  return eval(kids[0]) - eval(kids[1]);
}
efunc(e_div){
  //countArgs(2)
  return eval(kids[0]) / eval(kids[1]);
}
efunc(e_sin){
  //countArgs(1)
  return sin(eval(kids[0]));
}
efunc(e_cos){
  //countArgs(1)
    return cos(eval(kids[0]));
}
efunc(e_atan){
  //countArgs(1)
    return atan(eval(kids[0]));
}
efunc(e_ifgeq){
  //countArgs(4)
    if (eval(kids[0]) >= eval(kids[1])){
      return eval(kids[2]);
    } else {
      return eval(kids[3]);
    }
}
efunc(e_ifleq){
  //countArgs(4)
    if (eval(kids[0]) <= eval(kids[1])){
      return eval(kids[2]);
    } else {
      return eval(kids[3]);
    }
}
efunc(e_max){
  double a,b;
  //countArgs(2)
  a = eval(kids[0]);
  b = eval(kids[1]);
  return(a>b?a:b);
}
efunc(e_min){
  double a,b;
  //countArgs(2)
  a = eval(kids[0]);
  b = eval(kids[1]);
  return(a<b?a:b);
}
efunc(e_exp){
  //countArgs(1)
    return exp(eval(kids[0]));
}

efunc(e_ang){
  return atan2(cy,cx);
}

efunc(e_mag){
  return sqrt( cy*cy + cx*cx);
}

efunc(e_dis){
  double a,b;
  a = eval(kids[0]);
  b = eval(kids[1]);
  return sqrt( (cx-a)*(cx-a) + (cy-b)*(cy-b) );
}

efunc(e_angdis){
  double a,b;
  a = eval(kids[0]);
  b = eval(kids[1]);
  return atan2( cy-b, cx-a);
}

/*
double (*(efns[]))( int type, double value, int nkids, node **kids, double cx, double cy, double cz, double ct ) = 
     {e_none, e_num, e_x, e_y, e_z, e_t, e_mult, e_add, e_sub, e_div, e_sin, e_cos, e_atan, e_ifgeq, e_ifleq, e_max, e_min};
*/

//double evaluate(node *n, double cx, double cy, double cz, double ct){
double evaluate(node *n){
  //return  (*(typelist[n->type].f))( n->type, n->value, n->nkids, n->kids, cx, cy, cz, ct) ; 
 return  (*(typelist[n->type].f))( n->type, n->value, n->nkids, n->kids) ;  
}

/*

char *t_names[] = {"none", "num", "x", "y", "z", "t", "*", "+", "-", "/", "sin", "cos", "atan", "ifgeq", "ifleq", "max", "min"};

  
  */
int treeSize( node *tree ){
  int tot=0;
  int i;
  for (i=0;i<tree->nkids;i++){
    tot += treeSize(tree->kids[i]);
  }
  return tot + 1;
}

void freeTree(node *tree){
  int i, n;
  n = tree->nkids;
  for( i=0; i<n; i++){
    freeTree(tree->kids[i]);
  }
  if( tree->nkids) free(tree->kids);
  free(tree);
}


// condenses branches not containing variables.
// doesn't affect original tree.
node *compileTree( node *root ){
  int i;
  int k;
  node *cnode;
  cnode = (node *)malloc(sizeof(node));
  cnode->type = root->type;
  cnode->value = root->value;
  cnode->nkids = root->nkids;
  cnode->kids = NULL;
  if (cnode->nkids){
    cnode->kids = (node **)malloc(sizeof(node *) * cnode->nkids);
    for(i=0;i<cnode->nkids; i++){
      cnode->kids[i] = compileTree( root->kids[i] );
    }
  }


  k = 0;

  // can do more complex optimizations for ifleq, ifgeq, but need to do this at the parent level.
  // so we look at children of the current node.  
  for( i=0; i<cnode->nkids; i++){
    node *n ;
    n = cnode->kids[i];
    if( n->type == T_IFLEQ ){
      if ( (n->kids[0])->type == T_NUM &&
	   (n->kids[1])->type == T_NUM    ) {
	if( (n->kids[0])->value <= (n->kids[1])->value ){
	  cnode->kids[i] = n->kids[2];
          freeTree(n->kids[3]);
	} else {
	  cnode->kids[i] = n->kids[3]; 
          freeTree(n->kids[2]);
	}
        freeTree(n->kids[0]);
	freeTree(n->kids[1]);
        free(n->kids);
        free(n);
      }
    }else if( n->type == T_IFGEQ ){
      if ( (n->kids[0])->type == T_NUM &&
	   (n->kids[1])->type == T_NUM    ) {
	if( (n->kids[0])->value >= (n->kids[1])->value ){
	  cnode->kids[i] = n->kids[2];
          freeTree(n->kids[3]);
	} else {
	  cnode->kids[i] = n->kids[3];
          freeTree(n->kids[2]);
	}
        freeTree(n->kids[0]);
	freeTree(n->kids[1]);
        free(n->kids);
        free(n);
      }
    }

    if ( (cnode->kids[i])->type != T_NUM ){
	k++;
    }
  }

  // all kids are numbers, so replace this node with its evaluation
  if( !k && cnode->nkids ){
    double r;
    r = evaluate(cnode);
    for( i=0; i<cnode->nkids; i++){
      freeTree(cnode->kids[i]);
    }
    free(cnode->kids);
    cnode->kids = NULL;
    cnode->nkids = 0;
    cnode->type = T_NUM;
    cnode->value = r;
  }

 
 
  return cnode;
  
}


// condenses branches not containing variables.
// doesn't affect original tree.
// This variant uses interval code for more info.
node *compileTreeI( node *root ){
  int i;
  int k;
  node *cnode;
  cnode = (node *)malloc(sizeof(node));
  cnode->type = root->type;
  cnode->value = root->value;
  cnode->nkids = root->nkids;
  cnode->kids = NULL;
  if (cnode->nkids){
    cnode->kids = (node **)malloc(sizeof(node *) * cnode->nkids);
    for(i=0;i<cnode->nkids; i++){
      cnode->kids[i] = compileTreeI( root->kids[i] );
    }
  }


  k = 0;


  // can do more complex optimizations for ifleq, ifgeq, max, min, but need to do this at the parent level.
  // so we look at children of the current node.  
  for( i=0; i<cnode->nkids; i++){
    node *n, *p ;
    n = cnode->kids[i];
    if( n->type == T_IFLEQ || n->type == T_IFGEQ || 
          n->type == T_MAX || n->type == T_MIN ){

       intervals *i1, *i2;
       int c;
       i1 = NULL;
       i2 = NULL;

 /* break for debugging */
 // if( n->nkids == 2 ){
 // if ( n->kids[0]->type == T_NUM && n->kids[1]->type == T_MULT && 
 //      n->type == T_MAX ){
 //   k = 0;
 // }
 // }


       i1 = ival( n->kids[0] );
       i2 = ival( n->kids[1] );
       c = compare_intervals(i1,i2);
       free_intervals(i1);
       free_intervals(i2);

       if( c == 0 ) {
         // results inconclusive
	 k++; // IMPORTANT!
         continue;
       }

       
       if( n->type == T_IFLEQ ){
	 if( c < 0 ){
	   p = n->kids[2];
           freeTree(n->kids[3]);
         } else {
           p = n->kids[3];
           freeTree(n->kids[2]);
         }
         freeTree(n->kids[0]);
         freeTree(n->kids[1]);
       } else if ( n->type == T_IFGEQ ) {
         if ( c > 0 ){
           p = n->kids[2];
           freeTree(n->kids[3]);
         } else {
           p = n->kids[3];
           freeTree(n->kids[2]);
         } 
         freeTree(n->kids[0]);
         freeTree(n->kids[1]);
       } else if ( n->type == T_MIN ){
         if( c < 0 ){
	   p = n->kids[0];
           freeTree(n->kids[1]);
         } else {
           p = n->kids[1];
           freeTree(n->kids[0]);
         } 
       } else if ( n->type == T_MAX ){
         if( c > 0 ){
	   p = n->kids[0];
           freeTree(n->kids[1]);
         } else {
           p = n->kids[1];
           freeTree(n->kids[0]);
         }
       }
       free(n->kids);
       free(n);
       cnode->kids[i] = p; 
    }    
       


    if ( (cnode->kids[i])->type != T_NUM ){
	k++;
    }
  }

  // all kids are numbers, so replace this node with its evaluation
  if( !k && cnode->nkids ){
    double r;
    r = evaluate(cnode);
    for( i=0; i<cnode->nkids; i++){
      freeTree(cnode->kids[i]);
    }
    free(cnode->kids);
    cnode->kids = NULL;
    cnode->nkids = 0;
    cnode->type = T_NUM;
    cnode->value = r;
  }

 
 
  return cnode;
  
}

void writeTree( FILE *fp, node *root ){
  int i; 
  int type = root->type;

  if ( type == T_NONE){
    fprintf(fp, " none ");
    return;
  } 
  if ( type == T_NUM ){
    fprintf(fp, " %.20g ", root->value);
    return;
  }
  if ( root->nkids == 0 ) {
    fprintf(fp, " %s ", typelist[type].name);
    return;
  }
  
  fprintf(fp, "%s(", typelist[type].name);
  for (i=0; i<root->nkids; i++ ){
    writeTree(fp,root->kids[i]);
    if ( i < root->nkids - 1) fprintf(fp, ",");
  }
  fprintf(fp, " )");
}

int isNumber(char *s){
  /*
  int i = 0;
  int dec = 0;
  int exp = 0;
  
  while( s[i] != '\0' ){
    if (s[i] == '.'){
      if( dec) {
        return 0;
      } else {
	dec = 1;
      }
    } else if ( !isdigit(s[i]) ){
      return 0;
    }
    i++;
  }
  return 1;
  */
  double r;
  int i;
  i = sscanf(s, "%lg", &r);
  if (i==EOF || i!=1) return 0;
  return 1;
}

nodetype *getNodeType(char *s){

  int i;

  if (isNumber( s )) {
    return &typelist[T_NUM];
  }

  for( i=0; i<TYPECOUNT; i++ ){
    if ( strcmp(s, typelist[i].name) == 0 ) {
      return &typelist[i];
    }
  }
  printf("unrecognized node type \"%s\"\n",s);
  exit(-1);
}

void readTree( FILE *fp, node **root ){
  char buf[256];
  int i=0;
  char c;
  int nkids = 0;
  nodetype *type;
  c = fgetc(fp);
  //  while ( c != EOF ) {
    i = 0;
    // skip leading spaces
    while( isspace(c) ) { c = fgetc(fp); }
    // get token
    while ( c != '(' && c != ')' && c != ' ' && c != ',' && !isspace(c) ){
      buf[i++] = c;
      c = fgetc(fp);
    }
    if ( c==')' || c ==',' ){
      ungetc(c,fp);
    }
    // null terminate string
    buf[i] = '\0';
    type = getNodeType(buf);
    *root = (node *)malloc(sizeof(node));
    (*root)->type = type->index;
    (*root)->nkids = type->nkids;
    if( type->index == T_NUM ) {
      (*root)->value = atof(buf);
    }
    if( type->nkids > 0 ) {
      nkids = type->nkids;
      (*root)->kids = (node **)malloc(sizeof(node *) * nkids );
      for( i=0; i<nkids; i++){
        readTree( fp, &((*root)->kids[i]));
        if (i == nkids-1){
	  c = fgetc(fp);
          while ( c !=  ')' ){
            // get trailing spaces
	    c = fgetc(fp);
          }
        } else {
          c = fgetc(fp);
          while ( c != ','){
            c = fgetc(fp);
          }
        }
      }
    }
    //}
}

int getRandomNodeType(){
  int type;
  int tot=0;
  int i;
  
  for (i=0;i<TYPECOUNT;i++){
    tot += typelist[i].weight;
  }

  type =((int)( ((double)rand()/(double)RAND_MAX)*tot)) % tot;
  
  tot = 0;

  // avoid giving type "none"
  if (type == 0) return getRandomNodeType();

  for (i=0;i<TYPECOUNT;i++){
    tot += typelist[i].weight;
    if (tot >= type){
      return i;
    }
  }
}

node *genRandomTree(){
  int type;
  int i;
  node *tree;
  type = getRandomNodeType();

  tree = (node *)malloc(sizeof(node));
  tree->type = type;
  tree->nkids = typelist[type].nkids;

  if( type == T_NUM ){
     tree->value = (double)RAND_RANGE * (double)rand()/(double)RAND_MAX;
  }

  if ( tree->nkids ){
    tree->kids = (node **)malloc(tree->nkids * sizeof(node *));
  } else {
    tree->kids = NULL;
  }

  for( i=0; i<tree->nkids; i++){
    tree->kids[i] = genRandomTree();
  }
  return tree;
}

/* returns 1 if trees are the same, 0 if different. */
int compareTrees( node *t1, node *t2){
  int i;
  if (t1->type != t2->type) return 0;
  if ((t1->type == T_NUM) && (t1->value != t2->value)) return 0;
  if (t1->nkids != t2->nkids) return 0;
  for( i=0; i<t1->nkids; i++){
    if (!compareTrees(t1->kids[i], t2->kids[i])) return 0;
  }
  return 1;
}


void genGreyImageN(node *tree, unsigned char *p, int w, int h){
 double px, py;
  int x,y;
  double r;
  double xoff;
  double yoff;
  unsigned char *p2;
  double max;
  double *scratch;
  double *s2;
  scratch = (double *)malloc(sizeof(double)*w*h);
  max = 0;
  s2 = scratch;
  xoff = (double)40.0/(double)w;
  yoff = (double)40.0/(double)h;
  py = -20;
  px = -20;
  p2 = p;
  for( y = 0; y<h; y++){
    py += yoff;
    px = -20;
    for(x=0; x<w; x++){
      px += xoff;
      r = evaluate(tree);
      // r = evaluate(tree, px, py, 0,0);
      *(s2++) = r;
      if (r > max) max = r;
    }
  }
  s2 = scratch;
  for ( y=0; y<w*h;y++){
    *(p2++) = (unsigned char)((*(s2++) / max) * 255);
  }
}

#define DISPLAY 1
/* returns max value of tree over region */
double genDoubleImage(node *tree, double *p, int pw, int ph, 
                      double w, double h, double tcx, double tcy){
  double xoff, yoff;
  double *p2;
  double max;
  double px, py;
  int x,y;

  node *ctree;

  clock_t t0, t1, t2, t3;
  t0 = clock();
  // add timers to compilation & evaluation steps.

  if( compile == 2 ){
   ctree = compileTreeI(tree);
  } else if ( compile == 1 ){
    ctree = compileTree(tree);
  } else {
    ctree = tree;
  }

  t1 = clock();

  //printf(" Tree \n");
  //writeTree(stdout,tree);
  //printf(" \n compiles to tree \n");
  //writeTree(stdout,ctree);
  //printf("\n");
  printf(" %d -> %d ", treeSize(tree),treeSize(ctree));

  xoff = w / (double)pw;
  yoff = h / (double)ph;
  px = tcx - w/2;
  py = tcy - h/2;
  max = 0;
  p2 = p;
  t2 = clock();
  for( y=0; y<ph; y++){
    for( x=0; x<pw; x++){

      //*p2 = evaluate(tree, px, py, 0, 0);

      cx=px;
      cy = py;
      //*p2 = evaluate(tree);
      *p2 = evaluate(ctree);

      if (*p2>max) max = *p2;
      p2++;
      px += xoff;
    }
    px = tcx - w/2;
    py+=yoff;
#ifdef DISPLAY
    if (!(y%8)) { printf("*"); fflush(stdout); }
#endif
  }

  t3 = clock();
  printf(" %gs Compile, %gs render.", (double)(t1-t0)/(double)CLOCKS_PER_SEC, (double)(t3-t2)/(double)CLOCKS_PER_SEC);
  
  if( compile ) freeTree(ctree); 

  return max;
}


makeNormalizedPPM( double *db, double max, unsigned char *pb, int size){
  int i;
  double n;
  long c;
  unsigned char r,g,b;
  for(i=0; i<size; i++){
    // normalize:
    //n = *db / max;
    //n *= 256*256*256;
    //c = (long)n;

    // or, do modulos...
    c = (long)( *db * 256*256*256) % (256*256*256);

    //    printf("%d ",c);
    

    // a note on the colour scheme:
    //  imagine an rgb cube, 256 units on a side.
    //  we move one unit at a time, like a snake, heading in the b direction
    //  until we hit the opposite wall, then going up in the g direction one unit
    //  and  turning around to go back in the b direction. repeat until hit the top
    //  of the g axis, then go in the r direction one unit, and repeat whole process.

    r = (unsigned char )((c & (255 << 16 ))>>16);
    g = (unsigned char)((c & (255 << 8 ))>>8);
    b = (unsigned char)(c & 255);
    if (r%2) g = 255 - g;
    if (g%2 || r%2 ) b = 255-b;

    *(pb++) = r;
    *(pb++) = g;
    *(pb++) = b;

    db++;

  }

}

void writePPM( char *name, unsigned char *p, int w, int h){
  FILE *fp;
  fp = fopen(name,"wb");
  fprintf(fp, "P6\n%d %d\n255\n", w, h);
  fwrite(p, sizeof(unsigned char), 3*w*h, fp);
  fclose(fp);
}

void writeTreePPM(char *name, node *tree, int pw, int ph ){
  double *dbuf;
  double max;
  unsigned char *pbuf;
  dbuf = (double *)malloc(pw*ph*sizeof(double));
  max = genDoubleImage(tree, dbuf, pw, ph, RANGE_X, RANGE_Y, CENTRE_X, CENTRE_Y);
  pbuf = (unsigned char *)malloc(pw*ph*3*sizeof(unsigned char));
  makeNormalizedPPM(dbuf, max, pbuf, pw * ph);
  free(dbuf); 
  writePPM( name, pbuf, pw, ph);
  free(pbuf);
}



void genGreyImage( node *tree, unsigned char *p, int w, int h ){
  double px, py;
  int x,y;
  double r;
  double xoff;
  double yoff;
  unsigned char *p2;


  xoff = (double)40.0/(double)w;
  yoff = (double)40.0/(double)h;
  py = -20;
  px = -20;
  p2 = p;
  for( y = 0; y<h; y++){
    py += yoff;
    px = 0;
    for(x=0; x<w; x++){
      px += xoff;
      //r = evaluate(tree, px, py, 0,0);
      r = evaluate(tree);
      *(p2++) = (unsigned char)( (int)( r * 255 ) % 256 );
    }
  }
}

void writePGM( char *name, unsigned char *p, int w, int h){
  FILE *fp;
  fp = fopen(name,"wb");
  fprintf(fp, "P5\n%d %d\n255\n", w, h);
  fwrite(p, sizeof(unsigned char), w*h, fp);
  fclose(fp);
}




void writeTreeImage( char *name, node *tree, int w, int h){
  unsigned char *buf;
  buf = (unsigned char *)malloc(sizeof(unsigned char) * w * h );
  genGreyImageN(tree, buf, w, h);
  writePGM( name, buf, w, h);
  free(buf);
}



void testio(){
  node *t1, *t2;
  FILE *fp;
  char fn[256];
  char cmd[1024];
  int c = 0;
  pid_t lastpid = 0;

  srand(time(NULL));
  while(1){
    t1 = genRandomTree();
    while( treeSize(t1) < 4){
      freeTree(t1);
      t1 = genRandomTree();
    }
    fp = fopen("scratch.tree", "w");
    writeTree(fp, t1);
    fclose(fp);
    fp = fopen("scratch.tree", "r");
    readTree( fp, &t2);
    fclose(fp);
     
    if (!compareTrees(t1,t2)){
      printf("ERROR: Wrote \n");
      writeTree(stdout,t1);
      printf("\n, read \n");
      writeTree(stdout,t2);
      printf("\n");
      exit(-1);
    } else {
      printf("Tree ");
      writeTree(stdout, t1);
      sprintf(fn, "%.10d.pgm", c);
      c++;
      writeTreePPM( fn, t1, RES_X, RES_Y);
      printf("\n is ok.\n");
    }
    
    freeTree(t1);
    freeTree(t2);
    cmd[0] = '\0';
    
    strcat(cmd, "xv ");
    strcat(cmd, fn);
    // kill last display if ready for new one
    if (lastpid){

      system("killall xv");
      kill(lastpid, SIGABRT);
    }
    lastpid = fork();
    if (!lastpid){
	system(cmd);
        exit(0);
    }


    //getchar();
  }
}

// depth first, left to right, not counting root node
node **getithnode(node *tree, int i){
  int j,k;
  if (i==1) { 

    if ( tree->nkids > 0 ) {
        return &(tree->kids[0]);
    } else {
      printf("not enough kids.\n");
      exit(-1);
    }    
 
  }

  
  for( j=0; j<tree->nkids; j++){
    if ( ( k = treeSize(tree->kids[j]) ) >= i )  {
      if( i == 1) {
        return &(tree->kids[j]); 
      } else {
	return getithnode(tree->kids[j],i-1);
      }
    } else {
      i -= k;
    }
  }
  printf("not enough kids 2.\n");
  exit (-1);
 
}


node *copyTree( node *t ){
  node *r;
  int i;
  r = (node *)malloc(sizeof(node));
  r->type = t->type;
  r->nkids = t->nkids;
  r->value = t->value;
  if( r->nkids ){
    r->kids = (node **) malloc(sizeof(node *) * r->nkids);
  } else {
    r->kids = NULL;
  }
  for( i=0; i < r->nkids; i++){
    r->kids[i] = copyTree( t->kids[i]);
  }
  return r;
}


eqn *copyEqn( eqn *e ){
  eqn *r;
  r= (eqn *)malloc(sizeof(eqn));
  r->tree = copyTree(e->tree);
  r->size = treeSize(r->tree);
  return r;

}

void freeEqn( eqn *e ){
  freeTree(e->tree);
  free(e);
}

// pick 2 random nodes, 1 from each eqn, and exchange them.
eqn *breed2( eqn *eqn1, eqn *eqn2){

  eqn *e1, *e2;
  int i1, i2;  
  node **n1, **n2;
  node *tmp;
  double d;

  // don't modify originals
  e1 = copyEqn(eqn1);
  e2 = copyEqn(eqn2);

  // -1 since don't count root node
  i1 =  (e1->size - 1) * ((double)rand()/(double)RAND_MAX);
  i2 = (e2->size - 1) * ((double)rand()/(double)RAND_MAX);

  if(i1 <1) i1 = 1;
  if(i2 <1) i2 = 1;

  n1 = getithnode(e1->tree, i1);
  n2 = getithnode(e2->tree, i2);
  
  //printf("e1: ");
  //writeTree(stdout,e1->tree);
  //printf("\ne2: ");
  //writeTree(stdout,e2->tree);
  //printf("\nSwap nodes %d in e1 and %d in e2.\n", i1,i2);

  tmp = *n2;
  *n2 = *n1;
  *n1 = tmp;

  //printf("e1 becomes: "); 
  //writeTree(stdout, e1->tree);
  //printf("\ne2 becomes: ");
  //writeTree(stdout, e2->tree);
  //printf("\n");

  e1->size = treeSize(e1->tree);
  e2->size = treeSize(e2->tree);

  // flip a coin to decide which to return
  d= (double)rand()/(double)RAND_MAX;
  if (d > 0.5) {
    freeEqn(e2);
    return e1;
  } else {
    freeEqn(e1);
    return e2;
  }
  

}




void mutate(eqn *e){
  int i;
  node **n;

    // -1 since don't count root node
  i =  (e->size - 1) * ((double)rand()/(double)RAND_MAX);


  if(i <1) i = 1;

  n = getithnode(e->tree, i);

  freeTree(*n);
  *n = genRandomTree();

  e->size = treeSize(e->tree);

  
}

/*
   at jonathan's suggestion, we mutate values of constants as well.
   this gives us a nicer amount of randomness.
 */
void mutatenums( node *tree ){
  double d;
  int i;
  if (tree->type == T_NUM ){
      d = (double)rand() / (double)RAND_MAX;
      if ( d > 0.9 ){
        tree->value += 0.1 * RAND_RANGE * ( (double)rand() / (double)RAND_MAX );
      }
  }
  for ( i=0; i<tree->nkids; i++){
    mutatenums(tree->kids[i]);
  }
}

// makes local copy of stud for safety.
void breed1( eqn **pop, int psize, eqn *stud ){
  int i,j;
  eqn *kid;
  eqn *mystud;
  mystud = copyEqn(stud);
  for( i=0; i<psize; i++){
    //printf("%d ", i); fflush(stdout);
    kid = breed2(mystud, pop[i]);
    freeEqn(pop[i]);
    pop[i] = kid;
    j = rand();
    if ( j > RAND_MAX /  2) {
      mutate( kid );
    }
    mutatenums( kid->tree );
  }
  freeEqn(mystud);
  
}

void evolve(eqn **pop, int psize){
  int b;
  int a,d;
  int c;
  char i;
  //srand((unsigned int)time(NULL));
  while(1){
  shuffle:
      a = (int)( psize * (double)rand()/(double)RAND_MAX) % psize;
      b = (int)( psize * (double)rand()/(double)RAND_MAX) % psize;
      c = (int)( psize * (double)rand()/(double)RAND_MAX) % psize;
      d = (int)( psize * (double)rand()/(double)RAND_MAX) % psize;
      printf("a:%d b:%d c:%d d:%d\n",a,b,c,d);
  display:
 #ifdef DISPLAY
      printf("A: ");
 #endif
      writeTreePPM( "a.ppm", pop[a]->tree, RES_X, RES_Y);
 #ifdef DISPLAY
      printf("\nB: ");
 #endif
      writeTreePPM( "b.ppm", pop[b]->tree, RES_X, RES_Y);
 #ifdef DISPLAY
      printf("\nC: ");
 #endif
      writeTreePPM( "c.ppm", pop[c]->tree, RES_X, RES_Y);
 #ifdef DISPLAY
      printf("\nD: ");
 #endif
      writeTreePPM( "d.ppm", pop[d]->tree, RES_X, RES_Y);   
  prompt:
      printf("\n?[abcd][s][p#][t#][m][b]\n");
  input:
      i = getchar();
      if( i == 'a' ){
	printf("stud = a %d\n", a); 
              writeTreePPM( "pick.ppm", pop[a]->tree, RES_X * 2, RES_Y * 2); 
	breed1(pop,psize,pop[a]);
        //mutate(pop[0]);
      } else if (i =='b'){
	printf("stud = b %d\n", b); 
      writeTreePPM( "pick.ppm", pop[b]->tree, RES_X * 2, RES_Y * 2);
	breed1(pop,psize,pop[b]);         
        //mutate(pop[0]);
      } else if (i == 'c'){
	printf("stud = c %d\n", c); 
    writeTreePPM( "pick.ppm", pop[c]->tree, RES_X * 2, RES_Y * 2);
	breed1(pop,psize,pop[c]);
        //mutate(pop[0]); 
      } else if (i == 'd'){
	printf("stud = d %d\n", d); 
    writeTreePPM( "pick.ppm", pop[d]->tree, RES_X * 2, RES_Y * 2);
	breed1(pop,psize,pop[d]);
        //mutate(pop[0]);     
      } else if (i == 's'){
	goto shuffle;
      } else if (i == 't'){
        int j;
        node *ctree;
        j = 0;
        scanf("%d",&j);
        if( j >= psize ) goto prompt;
        printf("(Compiled) Tree %d: ", j);
        if( compile == 2 ){
	  ctree = compileTreeI(pop[j]->tree);
        } else {
          ctree = compileTree(pop[j]->tree);
        }
	writeTree( stdout,ctree );
        freeTree(ctree);
        printf("\n");
        goto prompt;
      } else if (i == 'p'){
        int j;
        j = 0;
        scanf("%d",&j);
        if( j >= psize ) goto prompt;
	  printf("Tree %d: ", j);
	writeTree( stdout,pop[j]->tree );
        printf("\n");
        goto prompt;
      } else if ( i == 'm' ){
        compile = (compile + 1) % 3;
        if( compile == 2 ) {
          printf("interval compile on.\n");
        } else if ( compile == 1){
          printf("simple compile on.\n");
        } else {
          printf("compile off.\n");
	  }
        goto display;
      } else if ( i == '\n' ){
	goto input;
      } else { goto prompt; }


  }
}


void renderTreeAndWritePPM( char *name, node *tree, int pw, int ph, int x, int y, int label ){
  double *dbuf;
  double max;
  unsigned char *pbuf;
  Image pic;

  dbuf = (double *)malloc(pw*ph*sizeof(double));
  max = genDoubleImage(tree, dbuf, pw, ph, RANGE_X, RANGE_Y, CENTRE_X, CENTRE_Y);
  pbuf = (unsigned char *)malloc(pw*ph*3*sizeof(unsigned char));
  makeNormalizedPPM(dbuf, max, pbuf, pw * ph);
  free(dbuf); 
  writePPM( name, pbuf, pw, ph);

  if (label >= 0)  writeLabel( pbuf, pw, ph, 0, 0, label);
  pic.w = pw;
  pic.h = ph;
  pic.size = pw*ph;
  pic.bpp = 3;
  pic.data = pbuf;
  pic.name = NULL;
  G_place_pic(&pic, &buf, x, y);  

  free(pbuf);
}

void evolveG(eqn **pop, int psize){
  int b;
  int a,d;
  int c;
  char i;

  //srand((unsigned int)time(NULL));
  while(1){
  shuffle:
      a = (int)( psize * (double)rand()/(double)RAND_MAX) % psize;
      b = (int)( psize * (double)rand()/(double)RAND_MAX) % psize;
      c = (int)( psize * (double)rand()/(double)RAND_MAX) % psize;
      d = (int)( psize * (double)rand()/(double)RAND_MAX) % psize;
      printf("a:%d b:%d c:%d d:%d\n",a,b,c,d);
  display:
      {
          int j;
          char fname[10];
          for( j = 0;  j < 48; j++){
                 sprintf(fname, "%.2d.ppm", j);
		 renderTreeAndWritePPM( fname, pop[j]->tree, RES_X, RES_Y, RES_X * (j %8), RES_Y * (j/8), j);
          }
      }

/*
 #ifdef DISPLAY
      printf("A: ");
 #endif
      renderTreeAndWritePPM( "a.ppm", pop[a]->tree, RES_X, RES_Y,
0, 0);
 #ifdef DISPLAY
      printf("\nB: ");
 #endif
      renderTreeAndWritePPM( "b.ppm", pop[b]->tree, RES_X, RES_Y, RES_X, 0);
 #ifdef DISPLAY
      printf("\nC: ");
 #endif
      renderTreeAndWritePPM( "c.ppm", pop[c]->tree, RES_X, RES_Y, 0, RES_Y);
 #ifdef DISPLAY
      printf("\nD: ");
 #endif
      renderTreeAndWritePPM( "d.ppm", pop[d]->tree, RES_X, RES_Y, RES_X, RES_Y);   
*/
    
  prompt:
      printf("\n?[abcd][##][s][p#][t#][m][e#]\n");
  input:
      processEvents(&buf);

      /*
      i =   (char)(((int)( ((double)rand()/(double)RAND_MAX)*4)) % 4);
      i = 'a' + i;     
      */
      i = getchar();

     processEvents(&buf);
      if( i == 'a' ){
	printf("stud = a %d\n", a); 
	// renderTreeAndWritePPM( "pick.ppm", pop[a]->tree, RES_X * 2, RES_Y * 2, RES_X * 2, 0); 
	breed1(pop,psize,pop[a]);
        //mutate(pop[0]);
      } else if (i =='b'){
	printf("stud = b %d\n", b); 
	//renderTreeAndWritePPM( "pick.ppm", pop[b]->tree, RES_X * 2, RES_Y * 2, RES_X * 2, 0); 
	breed1(pop,psize,pop[b]);         
        //mutate(pop[0]);
      } else if (i == 'c'){
	printf("stud = c %d\n", c); 
	//renderTreeAndWritePPM( "pick.ppm", pop[c]->tree, RES_X * 2, RES_Y * 2, RES_X * 2, 0); 
	breed1(pop,psize,pop[c]);
        //mutate(pop[0]); 
      } else if (i == 'd'){
	printf("stud = d %d\n", d); 
	//renderTreeAndWritePPM( "pick.ppm", pop[d]->tree, RES_X * 2, RES_Y * 2, RES_X * 2, 0); 
	breed1(pop,psize,pop[d]);
        //mutate(pop[0]);
      } else if (isdigit(i)) {
	  char idi_c;  int idi_i;
          idi_c = getchar();	     
          if(isdigit(idi_c)){
	          idi_i = (i - '0') * 10 +  idi_c - '0';
          } else {
	 	idi_i = i - '0';
          }
          if (idi_i > psize -1 ) {
               printf("population size is only %d!\n", psize);
		goto prompt;
               
          }
          printf("stud = %d\n", idi_i);
          breed1(pop,psize,pop[idi_i]);
      } else if (i == 's'){
	goto shuffle;
      } else if (i == 't'){
        int j;
        node *ctree;
        j = 0;
        scanf("%d",&j);
        if( j >= psize ) goto prompt;
        printf("(Compiled) Tree %d: ", j);
        if( compile == 2 ){
	  ctree = compileTreeI(pop[j]->tree);
        } else {
          ctree = compileTree(pop[j]->tree);
        }
	writeTree( stdout,ctree );
        freeTree(ctree);
        printf("\n");
        goto prompt;
      } else if (i == 'p'){
        int j;
        j = 0;
        scanf("%d",&j);
        if( j >= psize ) goto prompt;
	  printf("Tree %d: ", j);
	writeTree( stdout,pop[j]->tree );
        printf("\n");
        goto prompt;
      } else if (i == 'e'){
        int j;
        j = 0;
        scanf("%d",&j);
        if( j >= psize ) goto prompt;
	printf("Blowing up tree %d: ", j);
        renderTreeAndWritePPM( "pick.ppm", pop[j]->tree, RES_X * 4, RES_Y * 4, RES_X * 8, 0, -1);
        goto prompt; 
       
      } else if ( i == 'm' ){
        compile = (compile + 1) % 3;
        if( compile == 2 ) {
          printf("interval compile on.\n");
        } else if ( compile == 1){
          printf("simple compile on.\n");
        } else {
          printf("compile off.\n");
	  }
        goto display;
      } else if ( i == '\n' ){
	goto input;
      } else { goto prompt; }


  }
}

int main(int argc, char **argv){
  /*
  FILE *in;
  FILE *out;q
  double r;
  node *tree;
  in = fopen("in.tree","r");
  readTree(in, &tree);
  fclose(in);
  out = fopen("out.tree","w");
  writeTree(out, tree);
  fclose(out);
  r = evaluate(tree, 1,0,0,0);
  printf("tree:\t");
  writeTree(stdout, tree);
  printf("\nresult:\t%g\n", r);
  */
  //testio();
  int i;
  eqn *pop[POPSIZE];


  srand((unsigned int)time(NULL));
  if(argc<2) {
    for(i=0;i<POPSIZE;i++){
      pop[i] = (eqn *)malloc(sizeof(eqn));
      pop[i]->tree =  genRandomTree();
      while( treeSize(pop[i]->tree) < 2 ){
	freeTree(pop[i]->tree);
	pop[i]->tree = genRandomTree();
      }
      pop[i]->size = treeSize(pop[i]->tree);
    
    }
  }else{
    FILE *fp;
    node *tree;
    fp = fopen(argv[1], "r");
    readTree(fp, &tree);
    fclose(fp);
    for(i=0;i<POPSIZE;i++){
      pop[i] = (eqn *)malloc(sizeof(eqn));
      pop[i]->tree = copyTree(tree);
      pop[i]->size = treeSize(pop[i]->tree);    
    }
  }


  G_init( RES_X * 12, RES_Y * 6 );
  buf.w = RES_X * 12;
  buf.h = RES_Y * 6;
  buf.size = buf.w * buf.h;
  buf.bpp = 3;
  buf.name = NULL;
  buf.data = (unsigned char *)malloc(sizeof(unsigned char) * buf.size *3);

  evolveG(pop,POPSIZE);

  G_end();
}
