#include <math.h>
#include <limits.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <sys/types.h>
#include <signal.h>
#include <time.h>
#include <string.h>
#include <unistd.h>

#include "types.h"
#define INFINITY INT_MAX
#define max(a,b) (a<b?b:a)
#define min(a,b) (a<b?a:b)

void free_intervals(intervals *i){
  int n;
  for( n=0; n<i->n; n++){
    free(i->is[n]);
  }
  free(i->is);
  free(i);
}

#define ci(i,x) check_interval(i,x)

interval *check_interval(interval *i, char *call){
  if( i->lo <= i->hi ){
    return i;
  } else {
    printf(" %g < %g ! (%s) \n", i->hi, i->lo , call );
    return i;
  }
}
 
interval *copy_interval(interval *i1){
  interval *r;
  r = (interval *)malloc(sizeof(interval));
  r->hi = i1->hi;
  r->lo = i1->lo;
  return r;
}

intervals *union_intervals(intervals *i1, intervals *i2){
  intervals *r;
  int i;
  r = (intervals *)malloc(sizeof(intervals));
  r->n = i1->n + i2->n;
  r->is = (interval **)malloc(sizeof(interval *) * r->n);
  for(i=0;i<i1->n; i++){
    (r->is)[i] = copy_interval(i1->is[i]);
  }
  for(i=0;i<i2->n; i++){
    (r->is)[i+i1->n] = copy_interval(i2->is[i]);
  }
  return r;
  
}
intervals *copy_intervals(intervals *i){
  intervals *r;
  int n;
  r = (intervals *)malloc(sizeof(intervals));
  r->n = i->n;
  r->is = (interval **)malloc(sizeof(interval *) * r->n);
  for(n=0;n<r->n;n++){
    ( r->is)[n] = copy_interval( (i->is)[n] );
  }
  return r;
}

intervals *clip_intervals( intervals *i, double lo, double hi){
  intervals *r;
  int c=0;
  int n,k;
  for(n=0;n<i->n;n++){
    if( i->is[n]->hi < lo || i->is[n]->lo > hi ) c++;
  }
  r = (intervals *)malloc(sizeof(intervals));
  r->n = i->n - c;
  r->is = (interval **)malloc(sizeof(intervals *) * r->n);
  k = 0;
  for(n=0; n<i->n; n++){
    if( !(i->is[n]->hi < lo || i->is[n]->lo > hi) ){
      r->is[k] = (interval *)malloc(sizeof(interval));
      r->is[k]->hi = min( hi, i->is[n]->hi );
      r->is[k]->lo = max( lo, i->is[n]->lo );
      k++;
    } 
  }
  return r;
}

double maxbound(intervals *i){
  int n;
  double max;
  max = i->is[0]->hi;
  for(n=1; n<i->n; n++){
    if(i->is[n]->hi > max) max = i->is[n]->hi;
  }
  return max;

}
double minbound(intervals *i){
  int n;
  double min;
  min = i->is[0]->lo;
  for(n=1; n<i->n; n++){
    if(i->is[n]->lo < min) min = i->is[n]->lo;
  }
  return min;

}
/*
   check this carefully...
 
 */
intervals *simplify_intervals(intervals *i){
  intervals *r, *t;
  interval **s;
  int j,k;
  int c;
  char *mask;

  c = 0;

  mask = (char *)malloc(sizeof(char) * i->n);
  t = copy_intervals(i);
  s = t->is;

  for(j=0;j<t->n;j++){
    mask[j] = 0;
  }

  for(j=0; j<t->n; j++){
    if( mask[j] ) continue;
    for(k=0;k<j; k++){
      if( mask[k] ) continue;
      if( ! ( (s[j]->lo > s[k]->hi) || (s[j]->hi < s[k]->lo))){
	// k & j intersect, so merge j into k and mark j as such.
        mask[j] = 1;
        c++;
        s[k]->hi = max( s[k]->hi, s[j]->hi);
        s[k]->lo = min( s[k]->lo, s[j]->lo);
        // restart the loop
        j = -1;
        break;
      }
    }
  }
  
  r = (intervals *)malloc(sizeof(intervals));
  r->n = t->n - c;
  r->is = (interval **)malloc(sizeof(interval *) * r->n);

  k=0;
  for( j=0; j<t->n; j++){
    if(mask[j]) continue;
    r->is[k++] = copy_interval( t->is[j] );
  }

  free_intervals(t);
  free(mask);

  return r;    
 
}


intervals *ival(node *n){
 return  (*(typelist[n->type].fi))( n->type, n->value, n->nkids, n->kids) ;  

}

#define ifunc(x) intervals * x( int type, double value, int nkids, node **kids)

interval *make_i( double lo, double hi ){
  interval *i;
  i = (interval *)malloc(sizeof(interval));
  i->hi = hi;
  i->lo = lo;
  return i;
}

intervals *make_is( double lo, double hi ){
  intervals *i;
  i = (intervals *)malloc(sizeof(intervals));
  i->n = 1;
  i->is = (interval **)malloc(sizeof(interval *));
  *(i->is) = make_i(lo,hi);
  return i;
}

interval *i_mult_2( interval *i1, interval *i2 ){
  interval *i;
  double hh, hl, lh, ll; 

  i = (interval *)malloc(sizeof(interval));

  hh = i1->hi * i2->hi;
  hl = i1->hi * i2->lo;
  lh = i1->lo * i2->hi;
  ll = i1->lo * i2->lo;

  i->hi = i->lo = hh;

  if( hl > i->hi ) i->hi = hl;
  if( hl < i->lo ) i->lo = hl;

  if( lh > i->hi ) i->hi = lh;
  if( lh < i->lo ) i->lo = lh;

  if( ll > i->hi ) i->hi = ll;
  if( ll < i->lo ) i->lo = ll;

  return i;

}

interval *i_add_2( interval *i1, interval *i2 ){
  interval *i;

  i = (interval *)malloc(sizeof(interval));

  i->lo = i1->lo + i2->lo;
  i->hi = i1->hi + i2->hi;
  return i;
}

interval *i_sub_2( interval *i1, interval *i2 ){
  interval *i;
  double hh, hl, lh, ll; 

  i = (interval *)malloc(sizeof(interval));

  hh = i1->hi - i2->hi;
  hl = i1->hi - i2->lo;
  lh = i1->lo - i2->hi;
  ll = i1->lo - i2->lo;

  i->hi = i->lo = hh;

  if( hl > i->hi ) i->hi = hl;
  if( hl < i->lo ) i->lo = hl;

  if( lh > i->hi ) i->hi = lh;
  if( lh < i->lo ) i->lo = lh;

  if( ll > i->hi ) i->hi = ll;
  if( ll < i->lo ) i->lo = ll;

  return i;

}

/* definitely wrong. fix!
   e.g. when i2 spans 0 .... 

   how do you handle infinity? 

 */
interval *i_div_2( interval *i1, interval *i2 ){
  interval *i;
  double hh, hl, lh, ll; 

  i = (interval *)malloc(sizeof(interval));


  /* attempt to hack something up to deal with div/0.
      and zero-spanning i2:
   */
  if ( i2->lo * i2->hi <= 0 ){
    i->hi = INFINITY;  
  } 

  if( (i2->lo == 0) &&( i2->hi == 0)){
    i->hi = i->lo = INFINITY;

  }else if ( i2->lo == 0 ){

    i->lo = i1->hi / i2->hi;
    lh = i1->lo / i2->hi;
    if( lh < i->lo ) i->lo = lh;

  } else if (i2->hi == 0 ){

    i->lo = i1->hi / i2->lo;
    ll = i1->lo / i2->lo;
    if( ll < i->lo ) i->lo = ll;

  } else {

    hh = i1->hi / i2->hi;
    hl = i1->hi / i2->lo;
    lh = i1->lo / i2->hi;
    ll = i1->lo / i2->lo;
  
    i->hi = i->lo = hh;

    if( hl > i->hi ) i->hi = hl;
    if( hl < i->lo ) i->lo = hl;

    if( lh > i->hi ) i->hi = lh;
    if( lh < i->lo ) i->lo = lh;

    if( ll > i->hi ) i->hi = ll;
    if( ll < i->lo ) i->lo = ll;
  }


  return i;

}


ifunc(i_none){ return make_is(0,0); }
ifunc(i_num){ return make_is(value,value); }
ifunc(i_x){ return make_is(CENTRE_X - RANGE_X/2, CENTRE_X + RANGE_X/2); }
ifunc(i_y){ return make_is(CENTRE_Y - RANGE_Y/2, CENTRE_Y + RANGE_Y/2); }
ifunc(i_mult){ 
  int nn, n1, n2;
  intervals *i1, *i2;
  intervals *i, *r;
  int j,k;
  i1 = ival( kids[0] );
  i2 = ival( kids[1] );
  n1 = i1->n;
  n2 = i2->n;
  nn = n1 * n2;
  i = (intervals *)malloc(sizeof(intervals));
  i->n = nn;
  i->is = (interval **)malloc(sizeof(interval *) * nn);
 
  for( j=0; j<n1; j++){
    for( k=0; k<n2; k++ ){
      i->is[ j*n2 + k ] = ci( i_mult_2( i1->is[j], i2->is[k]), "mult");
    }
  }
 free_intervals(i1);
 free_intervals(i2);
  r = simplify_intervals(i); // THINK ABOUT THiS...
  free_intervals(i);
  return r;
}

ifunc(i_add ){ 
  int nn, n1, n2;
  intervals *i1, *i2;
  intervals *i, *r;
  int j,k;
  i1 = ival( kids[0] );
  i2 = ival( kids[1] );
  n1 = i1->n;
  n2 = i2->n;
  nn = n1 * n2;
  i = (intervals *)malloc(sizeof(intervals));
  i->n = nn;
  i->is = (interval **)malloc(sizeof(interval *) * nn);
 
  for( j=0; j<n1; j++){
    for( k=0; k<n2; k++ ){
      i->is[ j*n2 + k ] = ci( i_add_2( i1->is[j], i2->is[k]), "add");
    }
  }
 free_intervals(i1);
 free_intervals(i2);
  r = simplify_intervals(i); // THINK ABOUT THiS...
  free_intervals(i);
  return r;
}


ifunc(i_sub ){ 
  int nn, n1, n2;
  intervals *i1, *i2;
  intervals *i, *r;
  int j,k;
  i1 = ival( kids[0] );
  i2 = ival( kids[1] );
  n1 = i1->n;
  n2 = i2->n;
  nn = n1 * n2;
  i = (intervals *)malloc(sizeof(intervals));
  i->n = nn;
  i->is = (interval **)malloc(sizeof(interval *) * nn);
 
  for( j=0; j<n1; j++){
    for( k=0; k<n2; k++ ){
      i->is[ j*n2 + k ] = ci( i_sub_2( i1->is[j], i2->is[k]), "sub");
    }
  }
  free_intervals(i1);
  free_intervals(i2);
  r = simplify_intervals(i); // THINK ABOUT THiS...
  free_intervals(i);
  return r;
}



ifunc(i_div ){ 
  int nn, n1, n2;
  intervals *i1, *i2;
  intervals *i, *r;
  int j,k;
  i1 = ival( kids[0] );
  i2 = ival( kids[1] );
  n1 = i1->n;
  n2 = i2->n;
  nn = n1 * n2;
  i = (intervals *)malloc(sizeof(intervals));
  i->n = nn;
  i->is = (interval **)malloc(sizeof(interval *) * nn);
 
  for( j=0; j<n1; j++){
    for( k=0; k<n2; k++ ){
      i->is[ j*n2 + k ] = ci(i_div_2( i1->is[j], i2->is[k]), "div");
    }
  }
  free_intervals(i1);
  free_intervals(i2);
  r = simplify_intervals(i); // THINK ABOUT THiS...
  free_intervals(i);
  return r;
}

// probably wrong
interval *i_sin_2(interval *i2){
  interval *r;
  interval *i;
  double ql,qh,f;
  i = copy_interval(i2);
  if( i->lo < 0 || i->hi < 0 ){
    f = max( fabs( i->lo - fmod( i->lo, M_PI * 2) ), 
	     fabs( i->hi - fmod(i->hi, M_PI *2)) )  + 10;
    i->lo += f;
    i->hi += f;
  }

  r = (interval *)malloc(sizeof(interval));
  if ( i->hi - i->lo >= M_PI * 2 ){
    r->lo = -1;
    r->hi = 1;
    return r;
  }
  if (i->hi == i->lo ){
    r->lo = r->hi = sin(i->lo);
    return r;
  }

  f = fmod( i->lo, M_PI * 2 );
  ql = f;
  qh = f + i->hi - i->lo;
  if ( ql <= M_PI / 2 ){
    if ( qh < M_PI/2 ){
      r->hi = sin( i->hi );
      r->lo = sin( i->lo );
    } else {
      r->hi = 1;
      if( qh >= 3 * M_PI / 2 ){
        r->lo = -1;
      } else {
        r->lo = min( sin(i->hi), sin(i->lo) );
      }
    }
  } else {
    if ( qh <= M_PI ){
      r->hi = sin(i->lo);
      r->lo = sin(i->hi);
    } else {
      r->lo = -1;
      if( qh >= 5 * M_PI /2 ) {
        r->hi = 1;
      } else {
	r->hi = max( sin(i->hi), sin(i->lo));
      }
    }
  }
  if( r->lo > r->hi ){
    printf("got (%g,%g) from (%g, %g). sin. \n", r->lo, r->hi, i->lo, i->hi);
  }

  free(i);
  return r; 
}

//probably wrong
interval *i_cos_2(interval *i2){
  interval *r;
  interval *i;
  double ql,qh,f;
  i = copy_interval(i2);
  if( i->lo < 0 || i->hi < 0 ){
    f = max( fabs( i->lo - fmod( i->lo, M_PI * 2) ), 
	     fabs( i->hi - fmod(i->hi, M_PI *2)) )  + 10;
    i->lo += f;
    i->hi += f;
  }

  

  // if (i->lo > i->hi ) printf(" %g < %g ! (in i_cos_2) \n", i->hi, i->lo);

  r = (interval *)malloc(sizeof(interval));
  if ( i->hi - i->lo >= M_PI * 2 ){
    r->lo = -1;
    r->hi = 1;
    return r;
  }

  if (i->hi == i->lo ){
    r->lo = r->hi = cos(i->lo);
    return r;
  }
  f = fmod( i->lo, M_PI * 2 );
  ql = f;
  qh = f + i->hi - i->lo;

  if ( qh <= M_PI ){
    r->hi = cos(i->lo);
    r->lo = cos(i->hi);
  } else if ( qh >= 2 * M_PI ){
    r->hi = 1;
    if( ql <= M_PI ){
      r->lo = -1;
    } else if( qh >= 3 * M_PI){
      r->lo = -1;
    } else {
      r->lo = min( cos(i->hi), cos(i->lo));
    }
  } else {
    if ( ql > M_PI ) {
      r->lo = cos(i->lo);
      r->hi = cos(i->hi);
    } else {
      r->lo = -1;
      r->hi = max(cos(i->hi), cos(i->lo));
    }
  }
 
  if( r->lo > r->hi ){
    printf("got (%g,%g) from (%g, %g). cos. \n", r->lo, r->hi, i->lo, i->hi);
  }

  free(i);

  return r;
  
}

ifunc(i_sin){
  int n1, nn;
  intervals *i1, *i, *r;
  i1 = ival(kids[0]);
  n1 = i1->n;
  i = (intervals *)malloc(sizeof(intervals));
  i->n = n1;
  i->is = (interval **)malloc(sizeof(interval *) * n1);
  for(nn=0; nn < n1; nn++){
    i->is[nn] = ci( i_sin_2(i1->is[nn]), "sin");
  }
  free_intervals(i1);
  r = simplify_intervals(i); // THINK ABOUT THiS...
  free_intervals(i);
  return r;  
  
}

ifunc(i_cos){
  int n1, nn;
  intervals *i1, *i, *r;
  i1 = ival(kids[0]);
  n1 = i1->n;
  i = (intervals *)malloc(sizeof(intervals));
  i->n = n1;
  i->is = (interval **)malloc(sizeof(interval *) * n1);
  for(nn=0; nn < n1; nn++){
    i->is[nn] = ci(i_cos_2(i1->is[nn]),"cos");
  }
  free_intervals(i1);
  r = simplify_intervals(i); // THINK ABOUT THiS...
  free_intervals(i);
  return r;  
  
}


interval *i_atan_2(interval *i){
  interval *r;
  r = (interval *)malloc(sizeof(interval));
  r->lo = atan(i->lo);
  r->hi = atan(i->hi);
  return r;
  
}

ifunc(i_atan){
  int n1, nn;
  intervals *i1, *i, *r;
  i1 = ival(kids[0]);
  n1 = i1->n;
  i = (intervals *)malloc(sizeof(intervals));
  i->n = n1;
  i->is = (interval **)malloc(sizeof(interval *) * n1);
  for(nn=0; nn < n1; nn++){
    i->is[nn] = ci(i_atan_2(i1->is[nn]), "atan");
  }
  free_intervals(i1);
  r = simplify_intervals(i); // THINK ABOUT THiS...
  free_intervals(i);
  return r;  
}

ifunc(i_ifgeq){
  intervals *i1, *i2, *r, *r2;
  int j,k;
  int n1, n2;
  int c1, c2, c3;
  i1 = ival(kids[0]);
  i2 = ival(kids[1]);
  n1 = i1->n;
  n2 = i2->n;
  c1 = c2 = c3 = 0;
  for( j=0; j<n1; j++){
    for( k=0; k<n2; k++){
      if( (i1->is[j])->hi < (i2->is[k])->lo ){
	c1++;
      } else if ( (i1->is[j])->lo >= (i2->is[k])->hi ) {
        c2++;
      } else {
        c3++;
      }
      if ((c1 && c2) || c3){
        // ambiguous intervals
        free_intervals(i1);
        free_intervals(i2);
        i1 = ival(kids[2]);
        i2 = ival(kids[3]);
        r2 = union_intervals(i1,i2);
        free_intervals(i1);
        free_intervals(i2);
        r = simplify_intervals(r2);
        free_intervals(r2);  
        return r;
      }
    }
  }
  free_intervals(i1);
  free_intervals(i2);
  if (c1) {
    r = ival( kids[3] );
  } else {
    r = ival( kids[4] );
  }
  return r;

}

ifunc(i_ifleq){
  intervals *i1, *i2, *r, *r2;
  int j,k;
  int n1, n2;
  int c1, c2, c3;
  i1 = ival(kids[0]);
  i2 = ival(kids[1]);
  n1 = i1->n;
  n2 = i2->n;
  c1 = c2 = c3 = 0;
  for( j=0; j<n1; j++){
    for( k=0; k<n2; k++){
      if( (i1->is[j])->hi < (i2->is[k])->lo ){
	c1++;
      } else if ( (i1->is[j])->lo >= (i2->is[k])->hi ) {
        c2++;
      } else {
        c3++;
      }
      if ((c1 && c2) || c3){
        // ambiguous intervals

        free_intervals(i1);
        free_intervals(i2);
        i1 = ival(kids[2]);
        i2 = ival(kids[3]);
        r2 = union_intervals(i1,i2);
        free_intervals(i1);
        free_intervals(i2);
        r = simplify_intervals(r2);
        free_intervals(r2);
 
        return r;
      }
    }
  }
  free_intervals(i1);
  free_intervals(i2);
  if (c2) {
    r = ival( kids[3] );
  } else {
    r = ival( kids[4] );
  }
  return r;

}


ifunc(i_max){
  intervals *i1, *i2, *r, *r2;
  int j,k;
  int n1, n2;
  int c1, c2, c3;
  double hi, lo;
  i1 = ival(kids[0]);
  i2 = ival(kids[1]);
  n1 = i1->n;
  n2 = i2->n;
  c1 = c2 = c3 = 0;
  for( j=0; j<n1; j++){
    for( k=0; k<n2; k++){
      if( (i1->is[j])->hi <= (i2->is[k])->lo ){
	c1++;
      } else if ( (i1->is[j])->lo >= (i2->is[k])->hi ) {
        c2++;
      } else {
        c3++;
      }
      if ((c1 && c2) || c3){
        // ambiguous intervals
        r = union_intervals(i1,i2);
        lo = max( minbound(i1), minbound(i2));
        hi = max( maxbound(i1), maxbound(i2));
        free_intervals(i1);
        free_intervals(i2);
        r2 = clip_intervals(r, lo, hi);  
        free_intervals(r);
        r = simplify_intervals(r2);
        free_intervals(r2);
        return r;
      }
    }
  }

  if (c2) {
    r = i1;
    free_intervals(i2);
  } else {
    r = i2;
    free_intervals(i1);
  }
  return r;

}

ifunc(i_min){
  intervals *i1, *i2, *r, *r2;
  int j,k;
  int n1, n2;
  int c1, c2, c3;
  double hi, lo;
  i1 = ival(kids[0]);
  i2 = ival(kids[1]);
  n1 = i1->n;
  n2 = i2->n;
  c1 = c2 = c3 = 0;
  for( j=0; j<n1; j++){
    for( k=0; k<n2; k++){
      if( (i1->is[j])->hi <= (i2->is[k])->lo ){
	c1++;
      } else if ( (i1->is[j])->lo >= (i2->is[k])->hi ) {
        c2++;
      } else {
        c3++;
      }
      if ((c1 && c2) || c3){
        // ambiguous intervals
        r = union_intervals(i1,i2);
        hi = min( maxbound(i1), maxbound(i2));
        lo = min( minbound(i1), minbound(i2));
        free_intervals(i1);
        free_intervals(i2);
        r2 = clip_intervals(r, lo, hi);
        free_intervals(r);
        r = simplify_intervals(r2);
        free_intervals(r2);
        return r;
      }
    }
  }

  if (c1) {
    r = i1;
    free_intervals(i2);
  } else {
    r = i2;
    free_intervals(i1);
  }
  return r;

}


int compare_intervals( intervals *i1, intervals *i2 ){
  int j,k;
  int n1, n2;
  int c1, c2, c3;
  n1 = i1->n;
  n2 = i2->n;
  c1 = c2 = c3 = 0;
  for( j=0; j<n1; j++){
    for( k=0; k<n2; k++){
      if( (i1->is[j])->hi <= (i2->is[k])->lo ){
	c1++;
      } else if ( (i1->is[j])->lo >= (i2->is[k])->hi ) {
        c2++;
      } else {
        c3++;
      }
      if ((c1 && c2) || c3){
        // ambiguous intervals
        return 0;
      }
    }
  }

  if (c1) {
    return -1;
  } else {
    return 1;
  }
  return 0;

}


interval *i_exp_2(interval *i){
  interval *r;
  r = (interval *)malloc(sizeof(interval));
  r->lo = exp(i->lo);
  r->hi = exp(i->hi);
  return r;
  
}

ifunc(i_exp){
  int n1, nn;
  intervals *i1, *i, *r;
  i1 = ival(kids[0]);
  n1 = i1->n;
  i = (intervals *)malloc(sizeof(intervals));
  i->n = n1;
  i->is = (interval **)malloc(sizeof(interval *) * n1);
  for(nn=0; nn < n1; nn++){
    i->is[nn] = ci( i_exp_2(i1->is[nn]), "exp");
  }
  free_intervals(i1);
  r = simplify_intervals(i); // THINK ABOUT THiS...
  free_intervals(i);
  return r;  
}

ifunc(i_ang){
  return make_is( -M_PI, M_PI);
}

ifunc(i_mag){
  return make_is( 0, sqrt( RANGE_X * RANGE_X + RANGE_Y*RANGE_Y) / 2.0 );

}


/* this is incorrect for now. lazy. may not keep this function. */
ifunc(i_dis){
  return make_is(0, RANGE_X * RANGE_X + RANGE_Y * RANGE_Y);
}

// this is overly loose. for now.
ifunc(i_angdis){
  return make_is( -M_PI, M_PI);

}
