#include "tree.h"
#include "intervals.h"


static int RAND_RANGE=50;
static int RANGE_X=50;
static int RANGE_Y=50;
static int CENTRE_X = 0;
static int CENTRE_Y = 0;
static int RES_X = 100;
static int RES_Y = 100;


// node types for tree

enum { T_NONE, 
       T_NUM,  // constant 
       T_X,    // x coord
       T_Y,    // y coord
       T_Z,    // z coord
       T_T,    // time, for animations
       T_MULT, // multiply
       T_ADD,  // add
       T_SUB,  // subtract
       T_DIV,  // divide (errors?)
       T_SIN,  // sine
       T_COS,  // cosine
       T_ATAN, // arctangent
       T_IFGEQ,// logic; if a >= b then c else d
       T_IFLEQ, // ditto, <=
       T_MAX,    // max(a,b); equiv to if a >= b then a else b
       T_MIN,     // min(a,b)
       T_EXP,   // exp(a)
       T_ANG,   // polar angle in (-Pi,Pi]
       T_MAG,   // polar magnitude
       T_DIS,   // sqrt( (x-a)^2 + (y-b)^2 )
       T_ANGDIS, // atan2( y-b, x-a)
       TYPECOUNT
} ;

typedef struct nodetype{
  int index;
  char *name;
  // double (*f)( int type, double value, int nkids, node **kids, double cx, double cy, double cz, double ct );
 double (*f)( int type, double value, int nkids, node **kids);
  intervals * (*fi)( int type, double value, int nkids, node **kids);
  int nkids;
  int weight;
} nodetype;


//#define efunc(x) double  x( int type, double value, int nkids, node **kids, double cx, double cy, double cz, double ct )
#define efunc(x) double  x( int type, double value, int nkids, node **kids )
efunc(e_none);
efunc(e_num);
efunc(e_x);
efunc(e_y);
efunc(e_z);
efunc(e_t);
efunc(e_mult);
efunc(e_add);
efunc(e_sub);
efunc(e_div);
efunc(e_sin);
efunc(e_cos);
efunc(e_atan);
efunc(e_ifgeq);
efunc(e_ifleq);
efunc(e_max);
efunc(e_min);
efunc(e_exp);
efunc(e_ang);
efunc(e_mag);
efunc(e_dis);
efunc(e_angdis);

#define ifunc(x) intervals * x( int type, double value, int nkids, node **kids)
ifunc(i_none);
ifunc(i_num);
ifunc(i_x);
ifunc(i_y);
ifunc(i_z);
ifunc(i_t);
ifunc(i_mult);
ifunc(i_add);
ifunc(i_sub);
ifunc(i_div);
ifunc(i_sin);
ifunc(i_cos);
ifunc(i_atan);
ifunc(i_ifgeq);
ifunc(i_ifleq);
ifunc(i_max);
ifunc(i_min);
ifunc(i_exp);
ifunc(i_ang);
ifunc(i_mag);
ifunc(i_dis);
ifunc(i_angdis);

static nodetype typelist[] = { { T_NONE, "none", e_none,i_none, 0, 0 },
			{ T_NUM, "num", e_num, i_num, 0, 9 },
			{ T_X, "x", e_x, i_x, 0, 5 },
			{ T_Y, "y", e_y, i_y, 0, 5},
			{ T_Z, "z", e_z, i_y, 0, 0}, /* wRITE THWSe i_z, i_t laTER MaB? */
			{ T_T, "t", e_t, i_y, 0, 0},
			{ T_MULT, "*", e_mult, i_mult,  2, 2},
			{ T_ADD, "+", e_add, i_add, 2, 2},
			{ T_SUB, "-", e_sub, i_sub, 2, 2},
			{ T_DIV, "/", e_div, i_div, 2, 2},
			{ T_SIN, "sin", e_sin, i_sin, 1, 2},
			{ T_COS, "cos", e_cos, i_cos, 1, 2},
			{ T_ATAN, "atan", e_atan, i_atan, 1, 2},
			{ T_IFGEQ, "ifgeq", e_ifgeq, i_ifgeq, 4, 2 },
			{ T_IFLEQ, "ifleq", e_ifleq, i_ifleq,  4, 2 },
			{ T_MAX, "max", e_max, i_max, 2, 2 },
			{ T_MIN, "min", e_min, i_min, 2, 2},
 		        { T_EXP, "exp", e_exp, i_exp, 1, 2},
			       {T_ANG, "ang", e_ang, i_ang, 0, 5},
			       {T_MAG, "mag", e_mag, i_mag, 0, 5},
			       {T_DIS, "dis", e_dis, i_dis, 2, 2},
			       {T_ANGDIS, "angdis", e_angdis, i_angdis, 2, 2}
};
